/*
 * mp3stat - scan mp3 for time and other bitrate info built with some code
 *           from mp3check.
 * Copyright (c) 2002 by Ed Sweetman <safemode@comcast.net>
 * mp3check - check mp3 file for consistency and print infos
 * Copyright (C) 1998 by Johannes Overmann <overmann@iname.com>
 */
#include <sys/stat.h>
#ifndef WIN32
#include <sys/mman.h>
#include <unistd.h>
#endif
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "ttypes.h"		// define uchar and such.
#include "mp3.h"		// mp3 related functions
#include "mp3info.h"		// mp3stat structure

// get header from pointer
inline              Header
get_header (const uchar * p)
{
  Header              h;
  uchar              *q = (uchar *) & h;
  q[0] = p[3];
  q[1] = p[2];
  q[2] = p[1];
  q[3] = p[0];
  return h;
}

// set header to pointer
inline void
set_header (uchar * p, Header h)
{
  uchar              *q = (uchar *) & h;
  p[0] = q[3];
  p[1] = q[2];
  p[2] = q[1];
  p[3] = q[0];
}

// return length of frame in bytes
inline int
frame_length (Header h)
{
  if (h.ID) {
    return ((((h.layer () == 1) ? 48000 : 144000) * h.bitrate ()) /
	    h.samp_int_rate ()) + h.padding_bit;
  } else {
    return ((((h.layer () == 1) ? 24000 : 72000) * h.bitrate ()) /
	    h.samp_int_rate ()) + h.padding_bit;
  }
}

//Detects the next valid header
inline int
find_next_header (const uchar * p, int len, int min_valid)
{
  int                 i;
  const uchar        *q = p;
  const uchar        *t;
  int                 rest, k, l;
  Header              h, h2;
  for (i = 0; i < len - 3; i++, q++) {
    if (*q == 255) {
      h = get_header (q);
      l = frame_length (h);
      if (h.isValid () && (l >= 21)) {
	t = q + l;
	rest = len - i - l;
	for (k = 1; (k < min_valid) && (rest >= 4); k++) {
	  h2 = get_header (t);
	  if (!h2.isValid ())
	    break;
	  if (!h2.sameConstant (h))
	    break;
	  l = frame_length (h2);
	  if (l < 21)
	    break;
	  t += l;
	  rest -= l;
	}
	if (k == min_valid)
	  return i;
      }
    }
  }
  return -1;			// not found
}

// Major function ...scans mp3 and gathers info as it does it
void
scan_mp3 (uchar * p, int len, statistic * mp3info2)
{
  int                 start = find_next_header (p, len, min_valid);
  int                 rest = len;
  int                 frame = 0;
  int                 l, s;
  double              frame_rate0 = 0;

  int                 tmp_counter1_m, tmp_counter2_m, tmp_counter3_m = 0;
  int                *rak_po_m[100];
  int                 rak_nr_m = 0;
  rak_po_m[rak_nr_m] = new int[10001];

  if (start >= 0) {
    // check whole file
    rest -= start;
    p += start;
    Header              head = get_header (p);
    l = frame_length (head);
    p += l;
    rest -= l;
    start += l;
    while (rest >= 4) {
      Header              h = get_header (p);
      frame_rate0 = h.bitrate ();
      mp3info2->total_frames++;
      mp3info2->total_bitrate += (frame_rate0 / 100);
      mp3info2->amnt_bitrates[place ((int) frame_rate0)] += 1;
      rak_po_m[rak_nr_m][frame] = (int) frame_rate0;
      frame++;
      if (frame == 10001) {
	rak_nr_m++;
	rak_po_m[rak_nr_m] = new int[10001];
	frame = 0;
      }
      if (!(h.isValid () && (frame_length (h) >= 21))) {
	// invalid header, search for next valid header
	s = find_next_header (p, rest, min_valid);
	if (s < 0)
	  break;		// error: junk at eof
	// skips invalid bytes
	p += s;
	rest -= s;
	start += s;
      } else {
	head = h;
	// skip to next frame
	l = frame_length (h);
	p += l;
	rest -= l;
	start += l;
      }
    }
    if (mp3info2->bitrates != NULL)
      delete[]mp3info2->bitrates;
    mp3info2->bitrates = new int[10001 * (rak_nr_m + 1) + 1];
    for (tmp_counter1_m = 0; tmp_counter1_m <= rak_nr_m; tmp_counter1_m++) {
      for (tmp_counter2_m = 0; tmp_counter2_m <= 10000; tmp_counter2_m++) {
	mp3info2->bitrates[tmp_counter3_m] =
	  rak_po_m[tmp_counter1_m][tmp_counter2_m];
	tmp_counter3_m++;
      }
      delete[]rak_po_m[tmp_counter1_m];
    }
    mp3info2->average_bitrate =
      (mp3info2->total_bitrate / mp3info2->total_frames) * 100000;
    mp3info2->time_seconds = (len * 8) / mp3info2->average_bitrate;
  }
}

#ifndef WIN32
//  mmaps mp3 calls scan_mp3 which returns filled status structure mp3info
void
mp3stat (statistic * mp3info2)
{
  //create stat accessor
  struct stat         buf;
  int                 len;
  int                 fd;
  uchar              *p;
  uchar              *free_p;
  stat (mp3info2->name, &buf);
  len = buf.st_size;
  mp3info2->size = (double) len / 1024;
  // open file
  fd = open (mp3info2->name, O_RDONLY);
  if (fd == -1) {
    perror ("open");
  }
  // mmap file
  p = (uchar *) mmap (0, len, PROT_READ, MAP_SHARED, fd, 0);
  free_p = p;			//this should be left here, scan_mp3 is modifying p
  if (p == (uchar *) - 1) {
    perror ("mmap");
  }
  scan_mp3 (p, len, mp3info2);
  // unmap file and close
  if (munmap ((char *) free_p, len)) {
    perror ("munmap");
  }
  close (fd);
}

#else

#define WIN32_LEAN_AND_MEAN
#include <windows.h>
//  mmaps mp3 calls scan_mp3 which returns filled status structure mp3info
void
mp3stat (statistic * mp3info2)
{
  //create stat accessor
  struct stat         buf;
  int                 len;
  HANDLE              file;
  HANDLE              mapping;
  uchar              *p;
  uchar              *free_p;

  stat (mp3info2->name, &buf);
  len = buf.st_size;
  mp3info2->size = (double) len / 1024;

  // open file
  file = CreateFile (mp3info2->name, GENERIC_READ, FILE_SHARE_READ, NULL,
		     OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);
  if (file == INVALID_HANDLE_VALUE) {
    perror ("open");
    return;
  }
  // mmap file
  mapping = CreateFileMapping (file, NULL, PAGE_READONLY, 0, len, NULL);
  if (mapping == NULL) {
    perror ("CreateMapping");
    CloseHandle (file);
    return;
  }

  p = (uchar *) MapViewOfFile (mapping, FILE_MAP_READ, 0, 0, 0);
  free_p = p;			//this should be left here, scan_mp3 is modifying p
  if (p == NULL) {
    perror ("mmap");
    CloseHandle (mapping);
    CloseHandle (file);
    return;
  }
  scan_mp3 (p, len, mp3info2);

  // unmap file and close
  if (!UnmapViewOfFile (p)) {
    perror ("munmap");
  }
  CloseHandle (mapping);
  CloseHandle (file);
}

#endif
